#include <osbind.h>
#include <mintbind.h>
#include <string.h>
#include <setjmp.h>
#include <unistd.h>
#include <support.h>
#include "patchlev.h"
#include "global.h"
#include "pipe.h"


long _stksize = 65536L;

extern void dispatch __PROTO((void));


/* in_tfork() -- used in "put-self-in-background" trick.  See below. */
static jmp_buf  tforkj;
static int in_tfork(arg)
int arg;
{
  /* wait for parent to die before we can longjmp back */
  while (getppid () > 1)
    sleep (1);
  longjmp (tforkj, 1);
  /*NOTREACHED*/
  return 0;
}


/* cleanup() -- free everything we might have alloced, and destroy all
   semaphores we might have created. */
static void cleanup(void)
{
  cleanup_config();
  cleanup_mem();
  /* We have to be holding the semaphore to destroy it */
  Psemaphore(2, DMN_SEMAPHORE, -1L);
  Psemaphore(1, DMN_SEMAPHORE, 0L);
#ifdef DEBUG
  Psemaphore(2, DEBUG_SEMAPHORE, -1L);
  Psemaphore(1, DEBUG_SEMAPHORE, 0L);
#endif /* DEBUG */
}


int main(void)
{
  Cconws("\r\nGlueSTiK\277 STiK emulator for MiNTnet\r\n"
	 "Network interface daemon\r\n"
	 "Version " GS_VERSION "\r\n"
	 "\275 1996 Scott Bigham\r\n");
#ifdef DEBUG
  Cconws("    Trace logging enabled\r\n");
#endif
#ifdef BLOCK_OPEN
  Cconws("    Non-blocking open disabled\r\n");
#endif

#ifdef DEBUG
  debug("s", "Starting GSDaemon v" GS_VERSION);
#endif

  if (load_config_file() < 0) {
    cleanup();
    return -1;
  }

#ifdef RT_DEBUG
  {
    char *s = do_getvstr(GS_TRACE);

    if (s && (!strcmp(s, "1") || !stricmp(s, "TRUE") || !stricmp(s, "ON"))) {
      debugflag = 1;
      /* Slightly tricky:  If DEBUG and RT_DEBUG are both on, the
	 previous occurrence of this wrote nothing, since debugflag
	 was initialized to zero; so write it again. */
      debug("s", "Starting GSDaemon v" GS_VERSION);
    }
  }
#endif

#ifdef DEBUG
  debug("s", "Config file loaded");
#endif

  if (!init_mem()) {
    cleanup();
    return -1;
  }
#ifdef DEBUG
  debug("s", "Alloc pool initialized");
#endif

  if (!init_net()) {
    cleanup();
    return -1;
  }
#ifdef DEBUG
  debug("s", "Network functions initialized");
#endif

  Cconws("\r\n");

  /* "put-self-in-background" trick lifted from syslogd */
#if 0	/* when MiNT gets a non-blocking fork(), this will work */
  if (fork())
    exit(0);
#else	/* until then, we use this */
  if (!setjmp(tforkj) && tfork(in_tfork, 0) >= 0)
    _exit (0);
#endif

  dispatch();
  return 0;
}
